;;; LinGO big grammar specific functions

(in-package :lkb)

(defun establish-linear-precedence (rule-fs)
   ;;;    A function which will order the features of a rule
   ;;;    to give (mother daughter1 ... daughtern)
   ;;;    
   ;;;  Modification - this must always give a feature
   ;;;  position for the mother - it can be NIL if
   ;;; necessary
  (let* ((mother NIL)
         (daughter1 (get-value-at-end-of rule-fs '(ARGS FIRST)))
         (daughter2 (get-value-at-end-of rule-fs '(ARGS REST FIRST)))
         (daughter3 (get-value-at-end-of rule-fs '(ARGS REST REST FIRST))))
    (declare (ignore mother))
    (unless (and daughter1 (not (eql daughter1 'no-way-through)))
      (cerror "Ignore it" "Rule without daughter"))
    (append (list nil '(ARGS FIRST))
            (if (and daughter2 (not (eql daughter2 'no-way-through)))
                (list '(ARGS REST FIRST)))
            (if (and daughter3 (not (eql daughter3 'no-way-through)))
                (if (and daughter2 (not (eql daughter2 'no-way-through)))
                    (list '(ARGS REST REST FIRST)))))))


(defun spelling-change-rule-p (rule)
;;; a function which is used to prevent the parser 
;;; trying to apply a rule which affects spelling and
;;; which should therefore only be applied by the morphology
;;; system.  
;;; Old test was for something which was a subtype of
;;; *morph-rule-type* - this tests for 
;;; < NEEDS-AFFIX > = +
;;; in the rule
  (let ((affix (get-dag-value (tdfs-indef 
                               (rule-full-fs rule)) 'needs-affix)))
    (and affix (equal (type-of-fs affix) '(plus)))))




(defun redundancy-rule-p (rule)
;;; a function which is used to prevent the parser 
;;; trying to apply a rule which is only used
;;; as a redundancy rule 
;;; this version tests for 
;;; < PRODUCTIVE > = -
;;; in the rule
  (let ((affix (get-dag-value 
                (tdfs-indef (rule-full-fs rule)) 'productive)))
    (and affix (equal (type-of-fs affix) '(minus)))))
             




;;; return true for types that shouldn't be displayed in type hierarchy
;;; window. Descendents (if any) will be displayed, i.e. non-displayed
;;; types are effectively spliced out

(defun hide-in-type-hierarchy-p (type-name)
  ;; starts with _, or ends with _[0-9][M]LE[0-9]
  ;; or contains "GLBTYPE"
   (and (symbolp type-name)
      (or
         ;; graphs are pretty unreadable without glbtypes in there as well
         (search "GLBTYPE" (symbol-name type-name))
         (eql (char (symbol-name type-name) 0) #\_)
         (let* ((name (symbol-name type-name))
                (end (length name))
                (cur (position #\_ name :from-end t)))
            ;; wish I had a regexp package available...
            (and cur
               (< (incf cur) end)
               (if (digit-char-p (char name cur)) (< (incf cur) end) t)
               (if (eql (char name cur) #\M) (< (incf cur) end) t)
               (if (eql (char name cur) #\L) (< (incf cur) end))
               (if (eql (char name cur) #\E) (<= (incf cur) end))
               (or (= cur end)
                   (and (digit-char-p (char name cur)) (= (incf cur) end))))))))


(defun make-orth-tdfs (orth)
  (let ((unifs nil)
        (tmp-orth-path *orth-path*))
    (for orth-value in (split-into-words orth)
         do
         (let ((opath (create-path-from-feature-list 
                       (append tmp-orth-path *list-head*))))
           (push (make-unification :lhs opath                    
                                   :rhs
                                   (make-u-value 
                                    :types (list orth-value)))
                 unifs)
           (setq tmp-orth-path (append tmp-orth-path *list-tail*))))
    (let ((indef (process-unifications unifs)))
      (when indef
        (setf indef (create-wffs indef))
        (make-tdfs :indef indef)))))


#|
(defparameter *infl-pos-record* nil)

(defun find-infl-pos (unifs orth-string sense-id)
  (declare (ignore orth-string))
  (let ((types
         (for unif in unifs
              filter
              (when (null (path-typed-feature-list (unification-lhs unif)))
		(car (u-value-types (unification-rhs unif)))))))
    (cond
     ((null types) 
      (format t 
              "~%Warning ~A doesn't specify any types, no affix position found"
              sense-id)
      nil)
     ((cdr types)
      (format t 
              "~%Warning ~A specifies multiple types, no affix position found"
              sense-id))
     (t
      (let* ((type (car types))
	     (res (assoc type *infl-pos-record*)))
	(if res (cdr res)
	  (progn
	    (eval-possible-leaf-type *leaf-types* type)
	    (let ((type-entry (get-type-entry type)))
	      (cond (type-entry 
		     (let ((pos
			    (extract-infl-pos-from-fs 
			     (tdfs-indef (type-tdfs type-entry)))))
		       (unless (or pos 
                                   (or (subtype-p type 'non_affix_bearing)
                                       (subtype-p type 'infl_lex_entry)))
			 (format t "~%No position identified for ~A" sense-id))
		       (push (cons type pos) *infl-pos-record*)
		       pos))
		    (t
		     (format t "~%Warning ~A specifies invalid type, no affix position found"
			     sense-id)
		     nil))))))))))



(defun extract-infl-pos-from-fs (fs)  
  (let ((current-path '(ARGS))
         (coindexed-position 
          (existing-dag-at-end-of fs '(--FINAL-ARG)))
        (position 1))
    (if coindexed-position
        (loop (let* ((next-path 
                      (append current-path '(FIRST)))
                     (new-pos 
                      (existing-dag-at-end-of 
                       fs next-path)))
                (unless new-pos
                       (return nil))
                (when (eq new-pos coindexed-position)
                  (return position))
                (incf position)
                (setf current-path 
                  (append current-path '(REST))))))))



|#

;; Assign priorities to parser tasks
(defun rule-priority (rule)
  (case (rule-id rule)
    (extradj_i 100)
    (extradj_s 150)
    (extracomp 200)
    (extrasubj_f 300)
    (extrasubj_i 300)
    (fillhead_non_wh 150)
    (fillhead_wh_r 150)
    (fillhead_wh_subj_r 150)
    (fillhead_wh_nr_f 150)
    (fillhead_wh_nr_i 150)
    (fillhead_rel 100)
    (hoptcomp 200)
    (rootgap_l 100)
    (rootgap_r 100)
    (n_n_cmpnd 250)
    (nadj_i 400)    
    (adjh_i 350)
    (mid_coord_np 800)
    (top_coord_np 700)
    (hcomp 800)
    (hadj_s 400)
    (bare_np 300)
    (fin_non_wh_rel 200)
    (inf_non_wh_rel 100)
    (vpellipsis_lr 100)
    (taglr 100)
    (vgering 100)
    (temp_np 300)
    (numadj_np 200)
    (appos 200)
    (imper 300)
    (sailr 300)
    (advadd 300)
    (passive 400)
    (intransng 200)
    (transng 400)
    (monthdet 400)
    (weekdaydet 400)
    (monthunsat 400)
    (attr_adj 400)
    (partitive 400)
    (NP_part_lr 400)
    (dative_lr 400)
    (otherwise 
     (if (get-lex-rule-entry (rule-id rule))
	 400
       500))))

(defun gen-rule-priority (rule)
  (rule-priority rule))

(defparameter *unlikely-le-types* '(DISC_ADV_MLE1 LETTER_NAME_LE
				    MEALTIME_WORD_LE NUMADJ_NOUN_WORD_LE
				    NP_ADV_LE SUBCONJ_INF_LE 
				    SUBCONJ_INF_3MLE2
				    BE_TH_COP_IS_LE BE_TH_COP_ARE_LE
				    BE_TH_COP_WAS_LE BE_TH_COP_WERE_LE
				    INT_VP_ADV_POST_WORD_LE 
				    FREEREL_PRO_NP_WORD_LE
				    MV_SORB_LE PDET_ONE_LE 
				    DISC_ADV_LIKE_LE PREP_CP_LE
				    MV_OBJ_EQUI_NON_TRANS_PRD_LE
				    FREEREL_PRO_NP_WORD_LE
                                    MV_SUBJ_EQUI_PRD_LE
                                    MV_OBJ_EQUI_PRD_LE
                                    COORD_C_MLE
                                    NP_ADV_LE
				    ORDINAL_ADJ_LE
				    TO_C_PROP_ELIDED_LE
				    TO_C_NONPROP_ELIDED_LE
				    ))
(defparameter *likely-le-types* '(COORD_C_LE DISC_ADV_LE 
				  QUASIMODAL_LE MV_POSS_LE
				  HOUR_WORD_LE DITRANS_PREP_LE
				  MV_EXPL_IT_SUBJ_LIKE_LE
                                  S_ADV_PRE_WORD_NOSPEC_LE
				  MORE_ADJ_LE MV_SUBJ_EQUI_LE
				  PROPER_LE MV_PREP_PARTICLE_NP_LE
				  WH_PRO_WORD_LE
                                  MV_EMPTY_PREP*_INTRANS_LE
                                  MV_EMPTY_PREP_INTRANS_LE
				  TO_C_NONPROP_LE
				  HOW_ABOUT_N_OR_P_MLE1
				  AND_NUM_LE
				  COMPLEMENTED_UNSPECIFIED_CARD_LE
				  ))

#|
RFD-July10-ERROR when doing the batch parse does not recognize mrecord-history!!!

(defun lex-priority (mrec)
  (let ((lex-type (dag-type 
		   (tdfs-indef 
		    (if (mrecord-history mrec)
			(mhistory-fs (car (mrecord-history mrec)))
		      (mrecord-fs mrec))))))
    (cond ((member lex-type *unlikely-le-types* :test #'eq)
	   200)
	  ((member lex-type *likely-le-types* :test #'eq) 
	    800)
	  (t 400))))

(defun gen-lex-priority (fs)
  (let ((lex-type (dag-type (tdfs-indef fs)))) 
    (cond ((member lex-type *unlikely-le-types* :test #'eq) -200)
	  ((member lex-type *likely-le-types* :test #'eq) 800)
	  (t 600))))

|#


(defun set-temporary-lexicon-filenames nil
  (let ((prefix
         (if (and (boundp '*grammar-version*) 
                  (stringp (eval '*grammar-version*)))
             ;; avoid warnings due to unbound variable on compilation
             (remove-if-not #'alphanumericp 
                            (eval '*grammar-version*))
           "biglex")))
    (setf *psorts-temp-file* 
      (make-pathname :name prefix 
                     :directory (pathname-directory (lkb-tmp-dir))))
    (setf *psorts-temp-index-file* 
      (make-pathname :name (concatenate 'string prefix "-index") 
                     :directory (pathname-directory (lkb-tmp-dir))))
    (setf *leaf-temp-file* 
      (make-pathname :name (concatenate 'string prefix "-rels")
                     :directory (pathname-directory (lkb-tmp-dir))))))




;;; Function to run when batch checking the lexicon

(defun lex-check-lingo (new-fs id)
  #|
  (unless (extract-infl-pos-from-fs (tdfs-indef new-fs))
  (format *lkb-background-stream* "~%No position identified for ~A" id))
  |#
  (when new-fs
    (let* ((inflbool 
           (existing-dag-at-end-of (tdfs-indef new-fs)
                                   '(inflected)))
          (type (and (dag-p inflbool) (dag-type inflbool))))
      (when type
        (when
            (or (eq type 'bool)
                (and (consp type) (eq (first type) 'bool)))
          (format *lkb-background-stream* "~%INFLECTED unset on ~A" id))))))


(setf *grammar-specific-batch-check-fn* #'lex-check-lingo)


;;; **************************************************************
;;; RFD-10JULY08
;;; This is the batch-parse overwrite which is like LKB's
;;; batch-parse but also records the parse time for
;;; each sentence individually
;;; **************************************************************
(defun rfd-parse-sentences (&optional input-file (output-file 'unspec) &key rest)
  (unless input-file
     (setq input-file (ask-user-for-existing-pathname "Sentence file?")))
  ;; if xml input assume SAF XML
  (when (file-xml-p input-file)
    (return-from parse-sentences
      (apply #'process-saf-file-sentences (cons input-file rest))))
  (when
     (and input-file
          (or (probe-file input-file)
              (progn
                (show-message-window
                 (format nil "Input file `~a' does not exist" input-file))
                (return-from parse-sentences))))
     (with-open-file (istream input-file :direction :input)
        (if (eq output-file 'unspec)
          (setq output-file
            (ask-user-for-new-pathname "Output file?" input-file))
          (if (equal input-file output-file)
              (progn
                (show-message-window
                 (format nil "Attempt to overwrite input file `~a'"input-file))
                (return-from parse-sentences))))
        (unless output-file (return-from parse-sentences))
        (let ((line (read-line istream nil 'eof)))
           (if (and output-file (not (eq output-file t)))
              (with-open-file (ostream output-file :direction :output
                               :if-exists :supersede
                               :if-does-not-exist :create)
                (rfd-batch-parse-sentences istream ostream line
                                       #'extract-fine-system-sentence))
             (rfd-batch-parse-sentences istream t line
                                    #'extract-fine-system-sentence))))))


(defparameter *do-something-with-parse* nil)
(defparameter *lex-ids-used* nil)
(defparameter *parse-input* nil)
(defparameter *ostream* nil)

;;;RFD: needed for timing each parse.
(defparameter *rfd-start-time* 0)
(defparameter *rfd-end-time* 0)
(defparameter *rfd-limit* 0)

(defun rfd-batch-parse-sentences (istream ostream raw-sentence
&optional access-fn)
 (setf *lex-ids-used* nil)
 (setf *unanalysed-tokens* nil)
  (clear-type-cache)
  (format t "~%;;; Parsing test file~%") (finish-output t)
  (let ((nsent 0)
        (edge-total 0)
        (parse-total 0)
        ;; ask for recycling of safe dags
        ;; NB lexical entries must not contain safe dags - so
expand-psort-entry
        ;; and friends must rebind *safe-not-to-copy-p* to nil
        (*print-right-margin* 300)
        (start-time (get-internal-real-time)))
    (unwind-protect
      (loop
        (when (eql raw-sentence 'eof)
          (format t "~%;;; Finished test file")
          (unless (fboundp *do-something-with-parse*)
            (when ostream
              (format ostream "~%;;; Total CPU time: ~A msecs~%"
                      (- (get-internal-real-time) start-time))
              (format ostream "~%;;; Mean edges: ~,2F~%"
                      (/ edge-total nsent))
              (format ostream "~%;;; Mean parses: ~,2F~%"
                      (/ parse-total nsent))))
          (lkb-beep)
          (return))

        (let ((interim-sentence (if access-fn (apply access-fn (list
raw-sentence))
                                    raw-sentence)))
           (let ((sentence (string-trim '(#\Space #\Tab) interim-sentence)))
             (unless (or (equal sentence "") (char= (elt sentence 0) #\;))
               (incf nsent)
               (unless (fboundp *do-something-with-parse*)
                 ;; if we're doing something else,
                 ;; let that function control the output
                 (when ostream
                   (format ostream "~A ~A " nsent interim-sentence)
                   (finish-output ostream)))
                (let* ((munged-string
                        (if (fboundp 'preprocess-sentence-string)
                            (preprocess-sentence-string sentence)
                          sentence))
                       (user-input (split-into-words munged-string))
                       (*dag-recycling-p* t))
                  (#-:gdebug
                   handler-case
                   #+:gdebug
                   progn
                      (progn
                        #+:pooling
                        (reset-pools #+:gdebug t)

                        ;;;RFD:Timing individual parse
                        (setf *rfd-start-time* (get-internal-real-time))

                        (parse user-input nil)

                        (setf *rfd-end-time* (get-internal-real-time))
			(setf *rfd-limit* +1)
						
                        (setf *sentence* sentence)
                        (setf *parse-input* user-input)
                        (setf *ostream* ostream)
                        (when (fboundp *do-something-with-parse*)
                          (funcall *do-something-with-parse*)))
                      #-:gdebug
                      (storage-condition (condition)
				;;; RFD-11July08: In case parse failed due to the condition,
			        ;;; we still want to know the time  it took for parse to realize
				;;; the condition?!
			      (setf *rfd-end-time* (get-internal-real-time))
			      (setf *rfd-limit* -1)
                              (format t "Memory allocation problem: ~A caused by ~A~%"condition raw-sentence)
		     )
                      #+(and (not :gdebug) :allegro)
                      (EXCL:INTERRUPT-SIGNAL () (error "interrupt-signal"))
                      #-:gdebug
                      (error (condition)
			     ;;; RFD-11July08: In case parse failed due to the condition,
			     ;;; we still want to know the time  it took for parse to realize
			     ;;; the condition?!	
			     (setf *rfd-end-time* (get-internal-real-time))	
			     (setf *rfd-limit* -1)
                             (format t  "Error: ~A caused by ~A~%"condition raw-sentence)
			))
                  (unless (fboundp *do-something-with-parse*)
                    ;; if we're doing something else,
                    ;; let that function control the output
                    (when ostream
                      (let ((n (length *parse-record*)))
                        (setf parse-total (+ parse-total n))
                        (setf edge-total (+ edge-total *edge-id*))
                        (format ostream "~A ~A ~A ~A~%"
                                       n
                                       ;;(edge-count)
                                       *edge-id*
                                       (- *rfd-end-time* *rfd-start-time*)
				       *rfd-limit*
                         )
                       )
                      (finish-output ostream))))))
           (loop for lex-id in (collect-expanded-lex-ids *lexicon*)
              do
              (pushnew lex-id *lex-ids-used*))
           (setq raw-sentence (read-line istream nil 'eof))))
      (clear-chart)))) ; prevent any recycled dags from hanging around



